@@ -1,6 +1,7 @@
 from __future__ import annotations
 
 import enum
+import typing
 
 from zigpy_znp.types.cstruct import CStruct
 
@@ -28,19 +29,17 @@ class TrailingBytes(Bytes):
     Bytes must occur at the very end of a parameter list for easy parsing.
     """
 
-    pass
-
 
 def serialize_list(objects) -> Bytes:
     return Bytes(b"".join([o.serialize() for o in objects]))
 
 
 class FixedIntType(int):
-    _signed = None
-    _size = None
+    _signed: bool
+    _size: int
 
     def __new__(cls, *args, **kwargs):
-        if cls._signed is None or cls._size is None:
+        if getattr(cls, "_signed", None) is None or getattr(cls, "_size", None) is None:
             raise TypeError(f"{cls} is abstract and cannot be created")
 
         instance = super().__new__(cls, *args, **kwargs)
@@ -58,7 +57,7 @@ class FixedIntType(int):
             cls._size = size
 
         if hex_repr:
-            fmt = f"0x{{:0{cls._size * 2}X}}"
+            fmt = f"0x{{:0{cls._size * 2}X}}"  # type:ignore[operator]
             cls.__str__ = cls.__repr__ = lambda self: fmt.format(self)
         elif hex_repr is not None and not hex_repr:
             cls.__str__ = super().__str__
@@ -83,7 +82,7 @@ class FixedIntType(int):
 
         r = cls.from_bytes(data[: cls._size], "little", signed=cls._signed)
         data = data[cls._size :]
-        return r, data
+        return typing.cast(FixedIntType, r), data
 
 
 class uint_t(FixedIntType, signed=False):
@@ -162,7 +161,7 @@ class ShortBytes(Bytes):
     _header = uint8_t
 
     def serialize(self) -> Bytes:
-        return self._header(len(self)).serialize() + self
+        return self._header(len(self)).serialize() + self  # type:ignore[return-value]
 
     @classmethod
     def deserialize(cls, data: bytes) -> tuple[Bytes, bytes]:
@@ -182,7 +181,7 @@ class BaseListType(list):
     @classmethod
     def _serialize_item(cls, item, *, align):
         if not isinstance(item, cls._item_type):
-            item = cls._item_type(item)
+            item = cls._item_type(item)  # type:ignore[misc]
 
         if issubclass(cls._item_type, CStruct):
             return item.serialize(align=align)
@@ -215,7 +214,7 @@ class LVList(BaseListType):
     def deserialize(cls, data: bytes, *, align=False) -> tuple[LVList, bytes]:
         length, data = cls._header.deserialize(data)
         r = cls()
-        for i in range(length):
+        for _i in range(length):
             item, data = cls._deserialize_item(data, align=align)
             r.append(item)
         return r, data
@@ -242,7 +241,7 @@ class FixedList(BaseListType):
     @classmethod
     def deserialize(cls, data: bytes, *, align=False) -> tuple[FixedList, bytes]:
         r = cls()
-        for i in range(cls._length):
+        for _i in range(cls._length):
             item, data = cls._deserialize_item(data, align=align)
             r.append(item)
         return r, data
@@ -271,7 +270,7 @@ def enum_flag_factory(int_type: FixedIntType) -> enum.Flag:
     appropriate methods but with only one non-Enum parent class.
     """
 
-    class _NewEnum(int_type, enum.Flag):
+    class _NewEnum(int_type, enum.Flag):  # type:ignore[misc,valid-type]
         # Rebind classmethods to our own class
         _missing_ = classmethod(enum.IntFlag._missing_.__func__)
         _create_pseudo_member_ = classmethod(
@@ -286,7 +285,7 @@ def enum_flag_factory(int_type: FixedIntType) -> enum.Flag:
         __rxor__ = enum.IntFlag.__rxor__
         __invert__ = enum.IntFlag.__invert__
 
-    return _NewEnum
+    return _NewEnum  # type:ignore[return-value]
 
 
 class enum_uint8(uint8_t, enum.Enum):
@@ -321,33 +320,33 @@ class enum_uint64(uint64_t, enum.Enum):
     pass
 
 
-class enum_flag_uint8(enum_flag_factory(uint8_t)):
+class enum_flag_uint8(enum_flag_factory(uint8_t)):  # type:ignore[misc]
     pass
 
 
-class enum_flag_uint16(enum_flag_factory(uint16_t)):
+class enum_flag_uint16(enum_flag_factory(uint16_t)):  # type:ignore[misc]
     pass
 
 
-class enum_flag_uint24(enum_flag_factory(uint24_t)):
+class enum_flag_uint24(enum_flag_factory(uint24_t)):  # type:ignore[misc]
     pass
 
 
-class enum_flag_uint32(enum_flag_factory(uint32_t)):
+class enum_flag_uint32(enum_flag_factory(uint32_t)):  # type:ignore[misc]
     pass
 
 
-class enum_flag_uint40(enum_flag_factory(uint40_t)):
+class enum_flag_uint40(enum_flag_factory(uint40_t)):  # type:ignore[misc]
     pass
 
 
-class enum_flag_uint48(enum_flag_factory(uint48_t)):
+class enum_flag_uint48(enum_flag_factory(uint48_t)):  # type:ignore[misc]
     pass
 
 
-class enum_flag_uint56(enum_flag_factory(uint56_t)):
+class enum_flag_uint56(enum_flag_factory(uint56_t)):  # type:ignore[misc]
     pass
 
 
-class enum_flag_uint64(enum_flag_factory(uint64_t)):
+class enum_flag_uint64(enum_flag_factory(uint64_t)):  # type:ignore[misc]
     pass
